.pathAttributes <- function () {
  s <- search()
  paths <- lapply(1:length(s),
                  function(i) attr(as.environment(i), "path"))
  paths[[length(s)]] <- system.file()
  m <- sapply(paths, is.null)
  paths[m] <- s[m]
  unlist(paths)
}

.unmask <- function(s) gsub("@", "", s)

.mask <- local({
  regexp <- paste("(", paste(LETTERS, collapse = "|"), ")", sep = "")
  function(s) gsub(regexp, "@\\1", s)        # aBcD -> a@Bc@D
})

.upgrade <- function(directory = attr(pos.to.env(pos), "path"),
                     pos = 2, warn = TRUE) {
  if (is.character(pos)) pos <- match(pos, search())
  if (file.access(directory, mode = 2) < 0)
    stop("Cannot upgrade ", directory)
  newName <- directory
  while(file.exists(newName))
    newName <- paste(newName, "BAK", sep = "_")
  file.rename(directory, newName)
  dir.create(directory)
  for (i in dir(file.path(newName, "data"), "\\.RData$", all.files = TRUE))
    file.copy(file.path(newName, "data", i),
              file.path(directory, .mask(sub("\\.RData$", "", i))))
  if (warn)
    warning("Upgraded DDP ", sQuote(directory), "; old DDP in ", sQuote(newName),
            call. = FALSE)
  setdiff(dir(directory, all.files = TRUE), c(".", ".."))
}

.attach <- function(directory, pos = 2, warn = FALSE, readonly = FALSE) {
  env <- attach(NULL, pos, basename(directory))
  attr(env, "path") <- directory
  attr(env, "readonly") <- readonly
  if (file.exists(directory)) {
    fils <- setdiff(dir(directory, all.files = TRUE), c(".", ".."))
    if (("DESCRIPTION" %in% fils) &    # if an object, would have been 'm@shed'
        (.mask("DESCRIPTION") != "DESCRIPTION")) # have they changed .mask?
      fils <- .upgrade(directory, warn = warn)  # Upgrade old ddp
    objs <- .unmask(fils)
    fils <- file.path(directory, fils)
    for(i in seq(along = objs))
      eval(substitute(delayedAssign(OBJECT, {
        load(file = FILE)
        get(OBJECT)
      }), list(OBJECT = objs[i], FILE = fils[i])),
           envir = env)
  } else if (warn)
    warning(paste("'.save' will create:", directory), call. = FALSE)
}

.lib.loc <- function() {
  if((lib.loc <- Sys.getenv("LIB_LOC")) != "") return(lib.loc)
  if((lib.loc <- Sys.getenv("R_USER"))  != "") return(lib.loc)
  getwd()
}

.remote.lib <- function(LIB, default) {
  if((lib <- Sys.getenv(LIB)) != "" &&
     file.exists(file.path(.lib.loc(), lib))) return(lib)
  default
}

Attach <- function(lib = ".R_Store", lib.loc = getwd(),
                   pos = 2, uniquely = TRUE, ...) {
  if(class((x <- substitute(lib))) == "name")
    lib <- deparse(x) else lib <- lib

  path <- file.path(lib.loc, lib)
  if(uniquely)
    if(any(m <- (.pathAttributes() == path)))
      for(j in rev(sort(which(m)))) detach(pos = j)
  .attach(directory = path, pos = pos, ...)
}

AttachData <- function(...) Attach(lib = .remote.lib("LIB_DATA", ".R_Data"),
                                   lib.loc = .lib.loc(), ...)

AttachUtils <- function(...) Attach(lib = .remote.lib("LIB_UTILITIES", ".R_Utils"),
                                    lib.loc = .lib.loc(), ...)

Save <- Store <- function(..., lib = ".R_Store", lib.loc = getwd(),
                          remove = TRUE) {
  if(class((.x <- substitute(lib))) == "name")
    lib <- deparse(.x) else lib <- lib
  nam <-
    sapply(match.call(expand.dots = FALSE)$...,
           function(x) switch(class(x),
                              name = deparse(x),
                              call = {
                                o <- eval(x, envir = parent.frame(n = 4))
                                if(!is.character(o)) stop("non-character name!")
                                o
                              },
                              character = x,
                              stop("garbled call to 'Store'")))
  if(length(nam) == 0) return()
  nam <- drop(sort(unique(nam)))
  path <- file.path(lib.loc, lib)
  if(file.exists(path)) {
    if(!file.info(path)$isdir) stop(path, " exists but is not a directory!")
  } else dir.create(path)
  for(n in nam) {
    no <- !exists(n, inherits = FALSE, envir = parent.frame())
    comm <-
      substitute({
        PRE
        save(list = N, file = FILE)
        POST
      },
                 list(N = n, FILE = file.path(path, .mask(n)),
                      PRE = if(no) call("assign", n, call("get", n)) else NULL,
                      POST = if(no) call("remove", list = n) else NULL))
                                        #      print(comm)
    eval.parent(comm)
  }
  pos <- if(any(m <- (.pathAttributes() == path))) {
    m <- which(m)[1]
    detach(pos = m)
    m
  } else 2
  .attach(path, pos = pos)
  if(remove) {
    o <- intersect(eval.parent(quote(objects(all = TRUE))), nam)
    if(length(o) > 0)
      eval.parent(substitute(remove(list = o), list(o = o)))
  }
}


SaveData <- StoreData <- function(...) {
  Call <- match.call()
  Call[[1]] <- as.name("Save")
  if(is.null(Call$lib)) Call$lib <- .remote.lib("LIB_DATA", ".R_Data")
  if(is.null(Call$lib.loc)) Call$lib.loc <- .lib.loc()
  eval.parent(Call)
}

SaveUtils <- StoreUtils <- function(...) {
  Call <- match.call()
  Call[[1]] <- as.name("Save")
  if(is.null(Call$lib)) Call$lib <- .remote.lib("LIB_UTILITIES", ".R_Utils")
  if(is.null(Call$lib.loc)) Call$lib.loc <- .lib.loc()
  eval.parent(Call)
}

Objects <- function(lib = ".R_Store", lib.loc = getwd()) {
  if(class((x <- substitute(lib))) == "name")
    lib <- deparse(x) else lib <- lib  
  if(!is.na(pos <- match((path <- file.path(lib.loc, lib)), .pathAttributes())))
    objects(pos = pos, all = TRUE) else {
      .attach(path, pos = 2)
      objects(2, all = TRUE)
    }
}

ObjectsData <- function(...) Objects(lib = .remote.lib("LIB_DATA", ".R_Data"),
                                     lib.loc = .lib.loc(), ...)

ObjectsUtils <- function(...) Objects(lib = .remote.lib("LIB_UTILITIES", ".R_Utils"),
                                      lib.loc = .lib.loc(), ...)

Remove <- function(..., lib = ".R_Store", lib.loc = getwd()) {
  if(class((.x <- substitute(lib))) == "name")
    lib <- deparse(.x) else lib <- lib
  nam <- sapply(match.call(expand.dots = FALSE)$...,
                function(x) switch(class(x),
                                   name = deparse(x),
                                   call = {
                                     o <- eval(x, envir = parent.frame(n = 4))
                                     if(!is.character(o)) stop("non-character name!")
                                     o
                                   },
                                   character = x,
                                   stop("garbled call to 'Store'")))
  if(length(nam) == 0) return()
  nam <- .mask(sort(unique(nam)))
  path <- file.path(lib.loc, lib)
  mnam <- intersect(nam, dir(path, all.files = TRUE))
  if(length(d <- setdiff(nam, mnam)) > 0)
    warning(gettextf("Object %s not found and hence not removed\n", format(.unmask(d))))
  file.remove(file.path(path, mnam))
  pos <- if(any(m <- (.pathAttributes() == path))) {
    m <- which(m)[1]
    detach(pos = m)
    m
  } else 2
  .attach(path, pos = pos)
}

RemoveData <- function(...) {
  Call <- match.call()
  Call[[1]] <- as.name("Remove")
  if(is.null(Call$lib)) Call$lib <- .remote.lib("LIB_DATA", ".R_Data")
  if(is.null(Call$lib.loc)) Call$lib.loc <- .lib.loc()
  eval.parent(Call)
}

RemoveUtils <- function(...) {
  Call <- match.call()
  Call[[1]] <- as.name("Remove")
  if(is.null(Call$lib)) Call$lib <- .remote.lib("LIB_UTILITIES", ".R_Utils")
  if(is.null(Call$lib.loc)) Call$lib.loc <- .lib.loc()
  eval.parent(Call)
}
